@@ -1,6 +1,6 @@
 // Copyright 2008-present Contributors to the OpenImageIO project.
 // SPDX-License-Identifier: BSD-3-Clause
-// https://github.com/OpenImageIO/oiio/blob/master/LICENSE.md
+// https://github.com/OpenImageIO/oiio
 
 /*
   A few bits here are based upon code from NVIDIA that was also released
@@ -37,6 +37,7 @@
 #include <typeinfo>
 #include <type_traits>
 
+#include <OpenImageIO/Imath.h>
 #include <OpenImageIO/span.h>
 #include <OpenImageIO/dassert.h>
 #include <OpenImageIO/oiioversion.h>
@@ -192,12 +193,15 @@ inline OIIO_HOSTDEVICE int pow2rounddown(int x) { return floor2(x); }
 
 
 
-/// Round value up to the next whole multiple.
-/// For example, round_to_multiple(7,10) returns 10.
+/// Round value up to the next whole multiple. For example,
+/// `round_to_multiple(10,10) == 10`, `round_to_multiple(17,10) == 20`, and
+/// `round_to_multiple(-17,10) == -10`.
 template <typename V, typename M>
 inline OIIO_HOSTDEVICE V round_to_multiple (V value, M multiple)
 {
-    return V (((value + V(multiple) - 1) / V(multiple)) * V(multiple));
+    if (value >= 0)
+        value += V(multiple) - 1;
+    return value - (value % V(multiple));
 }
 
 
@@ -216,6 +220,20 @@ round_to_multiple_of_pow2(T x, T m)
 
 
 
+/// Round `value` down to a whole multiple of `multiple`. For example,
+/// `round_down_to_multiple(10,10) == 10`,
+/// `round_down_to_multiple(17,10) == 10`, and
+/// `round_down_to_multiple(-17,10) == -20`.
+template <typename V, typename M>
+inline OIIO_HOSTDEVICE V round_down_to_multiple (V value, M multiple)
+{
+    if (value < 0)
+        value -= V(multiple - 1);
+    return value - (value % V(multiple));
+}
+
+
+
 /// Multiply two unsigned 32-bit ints safely, carefully checking for
 /// overflow, and clamping to uint32_t's maximum value.
 inline OIIO_HOSTDEVICE uint32_t
@@ -750,28 +768,28 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE OUT_TYPE bit_cast (const IN_TYPE& in) {
     // about the memcpy in the general case.
     // FIXME: The intrinsics are not in clang <= 9 nor gcc <= 9.1. Check
     // future releases.
-    template<> OIIO_FORCEINLINE uint32_t bit_cast<float, uint32_t> (const float val) {
+    template<> OIIO_FORCEINLINE uint32_t bit_cast<float, uint32_t>(const float& val) {
           return static_cast<uint32_t>(_castf32_u32(val));
     }
-    template<> OIIO_FORCEINLINE int32_t bit_cast<float, int32_t> (const float val) {
+    template<> OIIO_FORCEINLINE int32_t bit_cast<float, int32_t>(const float& val) {
           return static_cast<int32_t>(_castf32_u32(val));
     }
-    template<> OIIO_FORCEINLINE float bit_cast<uint32_t, float> (const uint32_t val) {
+    template<> OIIO_FORCEINLINE float bit_cast<uint32_t, float>(const uint32_t& val) {
           return _castu32_f32(val);
     }
-    template<> OIIO_FORCEINLINE float bit_cast<int32_t, float> (const int32_t val) {
+    template<> OIIO_FORCEINLINE float bit_cast<int32_t, float>(const int32_t& val) {
           return _castu32_f32(val);
     }
-    template<> OIIO_FORCEINLINE uint64_t bit_cast<double, uint64_t> (const double val) {
+    template<> OIIO_FORCEINLINE uint64_t bit_cast<double, uint64_t>(const double& val) {
           return static_cast<uint64_t>(_castf64_u64(val));
     }
-    template<> OIIO_FORCEINLINE int64_t bit_cast<double, int64_t> (const double val) {
+    template<> OIIO_FORCEINLINE int64_t bit_cast<double, int64_t>(const double& val) {
           return static_cast<int64_t>(_castf64_u64(val));
     }
-    template<> OIIO_FORCEINLINE double bit_cast<uint64_t, double> (const uint64_t val) {
+    template<> OIIO_FORCEINLINE double bit_cast<uint64_t, double>(const uint64_t& val) {
           return _castu64_f64(val);
     }
-    template<> OIIO_FORCEINLINE double bit_cast<int64_t, double> (const int64_t val) {
+    template<> OIIO_FORCEINLINE double bit_cast<int64_t, double>(const int64_t& val) {
           return _castu64_f64(val);
     }
 #endif
@@ -809,6 +827,82 @@ swap_endian (T *f, int len=1)
 }
 
 
+#if (OIIO_GNUC_VERSION || OIIO_ANY_CLANG || OIIO_INTEL_CLASSIC_COMPILER_VERSION) && !defined(__CUDACC__)
+// CPU gcc and compatible can use these intrinsics, 8-15x faster
+
+template<> inline void swap_endian(uint16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap16(f[i]);
+}
+
+template<> inline void swap_endian(uint32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap32(f[i]);
+}
+
+template<> inline void swap_endian(uint64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap64(f[i]);
+}
+
+template<> inline void swap_endian(int16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap16(f[i]);
+}
+
+template<> inline void swap_endian(int32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap32(f[i]);
+}
+
+template<> inline void swap_endian(int64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = __builtin_bswap64(f[i]);
+}
+
+template<> inline void swap_endian(float* f, int len) {
+    swap_endian((uint32_t*)f, len);
+}
+
+template<> inline void swap_endian(double* f, int len) {
+    swap_endian((uint64_t*)f, len);
+}
+
+#elif defined(_MSC_VER) && !defined(__CUDACC__)
+// CPU MSVS can use these intrinsics
+
+template<> inline void swap_endian(uint16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ushort(f[i]);
+}
+
+template<> inline void swap_endian(uint32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ulong(f[i]);
+}
+
+template<> inline void swap_endian(uint64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_uint64(f[i]);
+}
+
+template<> inline void swap_endian(int16_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ushort(f[i]);
+}
+
+template<> inline void swap_endian(int32_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_ulong(f[i]);
+}
+
+template<> inline void swap_endian(int64_t* f, int len) {
+    for (int i = 0; i < len; ++i)
+        f[i] = _byteswap_uint64(f[i]);
+}
+#endif
+
+
 
 // big_enough_float<T>::float_t is a floating-point type big enough to
 // handle the range and precision of a <T>. It's a float, unless T is big.
@@ -948,7 +1042,7 @@ inline void convert_type<uint16_t,float> (const uint16_t *src,
 }
 
 
-#ifdef _HALF_H_
+#if defined(_HALF_H_) || defined(IMATH_HALF_H_)
 template<>
 inline void convert_type<half,float> (const half *src,
                                       float *dst, size_t n,
@@ -1016,7 +1110,7 @@ convert_type<float,uint8_t> (const float *src, uint8_t *dst, size_t n,
 }
 
 
-#ifdef _HALF_H_
+#if defined(_HALF_H_) || defined(IMATH_HALF_H_)
 template<>
 inline void
 convert_type<float,half> (const float *src, half *dst, size_t n,
@@ -1811,7 +1905,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_atan2 (float y, float x) {
 template<typename T>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_log2 (const T& xval) {
     using namespace simd;
-    typedef typename T::int_t intN;
+    typedef typename T::vint_t intN;
     // See float fast_log2 for explanations
     T x = clamp (xval, T(std::numeric_limits<float>::min()), T(std::numeric_limits<float>::max()));
     intN bits = bitcast_to_int(x);
@@ -1919,7 +2013,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_log1p (float x) {
 template<typename T>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_exp2 (const T& xval) {
     using namespace simd;
-    typedef typename T::int_t intN;
+    typedef typename T::vint_t intN;
 #if OIIO_SIMD_SSE
     // See float specialization for explanations
     T x = clamp (xval, T(-126.0f), T(126.0f));
@@ -1951,7 +2045,7 @@ OIIO_FORCEINLINE OIIO_HOSTDEVICE T fast_exp2 (const T& xval) {
 
 template<>
 OIIO_FORCEINLINE OIIO_HOSTDEVICE float fast_exp2 (const float& xval) {
-#if OIIO_NON_INTEL_CLANG && OIIO_FMATH_SIMD_FRIENDLY
+#if OIIO_ANY_CLANG && !OIIO_INTEL_LLVM_COMPILER && OIIO_FMATH_SIMD_FRIENDLY
     // Clang was unhappy using the bitcast/memcpy/reinter_cast/union inside
     // an explicit SIMD loop, so revert to calling the standard version.
     return std::exp2(xval);
